/*
 * ArgTree.java
 *
 * Created on 02 September 2008, 17:41
 * Represents the Argumentation Tree, and is reponsible for:
 *  1) adding and deleting nodes to/from this tree
 *  2) adding round moves to the tree at the end of each round
 *  3) recoloring the tree at the end of each round
 *  4) pruning player's p nodes from the tree when this player leaves the game
 *  5) providing the basic tree infromation in a compressed format
 * @author maya
 */

package GUIComponents;
//java packages
import java.io.*;
import java.util.*;

public class ArgTree {
    private ArgTreeNode[] leafs=null;//the leafs of the tree
    private ArgTreeNode root=null;//the root of the tree    
    private ArgTreeNode greenConfNode=null;//the node with the highest confidence on the tree
    private ArgTreeNode activeGreenConfNode=null;//the green node with the highest confidence on the tree
    private int treeSize=0; //the number of nodes in the current tree
    private int maxTreeSize;//the maximum number of the nodes allowed for this tree
    private int currentNoPlayers=0;//the current number of players in the game
    //private FocusingUnit focusingUnit=null;
    //private int focusingLevel=1;
    /***************************************************************************************************************/
    
    //Constructor
    public ArgTree(int currentNoPlayers, int maxTreeSize) {
        this.leafs = new ArgTreeNode[1];
        this.maxTreeSize = maxTreeSize;
        this.currentNoPlayers = currentNoPlayers;
    }
    /***************************************************************************************************************/        
    //Setters and Getters   
    protected ArgTreeNode[] getLeafs(){return this.leafs;}
    protected double getGreenConf(){
        if(this.greenConfNode==null) return -1;
        else    return this.greenConfNode.getConfidence();
    }
    protected double getActiveGreenConf(){
        if(this.activeGreenConfNode==null)  return -1;
        else    return this.activeGreenConfNode.getConfidence();
    }
    protected int getTreeSize(){return this.treeSize;}       
    protected void setCurrentNoPlayers(int newCurrentNoPlayers){this.currentNoPlayers = newCurrentNoPlayers;}
    protected int getCurrentNoPlayers(){return this.currentNoPlayers;}
   /* protected int getFocsuingLevel(){return this.focusingLevel;}
    protected void setFocusingLevel(int newFocusingLevel){this.focusingLevel = newFocusingLevel;}*/
    protected boolean rootAttacked(){
        if(root==null) return false;
        if(this.root.getAttacks()!=null) return true;
        else return false;
    }
    /**returns a copy of the tree root*/
    protected ArgTreeNode copyRoot(){
        if(this.root==null) return null;
        else return new ArgTreeNode(this.root);
    }
    /**returns true if the tree has any green leafs*/
    protected boolean hasGreenLeafs(){        
        for(int i=0; i<this.leafs.length; i++){
            if(this.leafs[i].getColour() == Colours.GREEN) return true;
        }
        return false;
    }
    /**returns true if the tree has any possible green leafs (purple leafs)*/
    private boolean hasPossibleGreenLeafs(){
        for(int i=0; i<this.leafs.length; i++){
            if(MoveType.originalColour(this.leafs[i].getType()) == Colours.GREEN) return true;
        }
        return false;
    }
    /***************************************************************************************************************/
     
    //Methods
    //Compressed Arg Tree Methods
    /** provides the tree leafs in compressed format*/
    protected CompressedArgTreeNode[] getCompressedLeafs(int focusingLevel){
        return this.getCompressedLeafs(focusingLevel, this.leafs);        
    }
    /** provides the tree leafs in compressed format*/
    protected CompressedArgTreeNode[] getCompressedLeafs(int focusingLevel,ArgTreeNode[] leafNodes){
        FocusingUnit focusingUnit = new FocusingUnit(this);
        switch (focusingLevel){
            case 1:
                
                ArgTreeNode[] strongestPosition = focusingUnit.strongestPosition();
                return this.compressedLeafs(ArgTreeNode.intersects(strongestPosition, leafNodes));
            case 2: case 3:
                ArgTreeNode focusedStrongestPosition = focusingUnit.focusedStrongestPositionRandom();
                if(focusedStrongestPosition.in(leafNodes)){
                    ArgTreeNode[] focusedStrongestPositionArray= new ArgTreeNode[1];
                    focusedStrongestPositionArray[0]=focusedStrongestPosition;
                    return this.compressedLeafs(focusedStrongestPositionArray);
                }
                else return null;
            default:
                return this.compressedLeafs(leafNodes);
        }
    } 
    
    /**provides the tree highest confidence leaf in compressed format*/
    protected CompressedArgTreeNode highestConfLeaf(int playerIndx){
        if(this.leafs==null)return null;
        CompressedArgTreeNode highestConfLeaf = null;        
        ArgTreeNode leaf = leafs[0];
        for(int i=1; i<leafs.length; i++){
            if(leafs[i]!=null){
                if((leafs[i].getPlayer()!=playerIndx)&&(leafs[i].getConfidence()>leaf.getConfidence()))
                    leaf = leafs[i];
            }                    
        }
        int player = leaf.getPlayer();
        int parent; 
        if (leaf.getParent()==null) parent =0;
        else    parent = leaf.getParent().getPlayer();
        highestConfLeaf = new CompressedArgTreeNode(player, parent, leaf.getLevel(), this.getLeafIndex(leaf));        
        return highestConfLeaf;
    }       
    /**returns the set leafNodes in proper compressed format*/
    protected CompressedArgTreeNode[] compressedLeafs(ArgTreeNode[] leafNodes){ 
        if(leafNodes==null) return null;
        CompressedArgTreeNode[] compressedleafs=new CompressedArgTreeNode[leafNodes.length];
        int index=0;
        for(int i=0; i<this.leafs.length; i++){
            if(leafs[i].in(leafNodes)){
                int player = leafs[i].getPlayer();
                int parent; 
                if (leafs[i].getParent()==null)
                    parent =0;
                else
                    parent = leafs[i].getParent().getPlayer();
                compressedleafs[index] = new CompressedArgTreeNode(player, parent, leafs[i].getLevel(), i);
                index++;

            }
        }
        return compressedleafs;
    }
     /***************************************************************************************************************/
    //LEAF METHODS
    /**returns the leafs the hasn't been attacked in this round*/
    private ArgTreeNode[] oldLeafs(boolean[] attacked){
        int length=0;
        ArgTreeNode[] oldLeafs = null;        
        for(int i=0; i<attacked.length; i++)
            if (!attacked[i])   length++;        
        if (length>0){
            oldLeafs = new ArgTreeNode[length];       
            int oldLeafIndx=0;
            for(int i=0; i<leafs.length; i++){
                if(!attacked[i]){
                    oldLeafs[oldLeafIndx]=leafs[i];
                    oldLeafIndx++;
                }
            }
        }                    
        return oldLeafs;        
    }   
    /**merges the new leafs (this round moves) with the unattacked leafs from previous rounds*/
    private void currentRoundLeafs(ArgTreeNode[] newLeafs,  boolean[] attacked){    
        ArgTreeNode[] oldLeafs = this.oldLeafs(attacked);
        int oldLeafsLength;
        if(oldLeafs ==null) {oldLeafsLength=0;}
        else {oldLeafsLength= oldLeafs.length;}
        int currentLeafsLength =  oldLeafsLength + newLeafs.length;
        ArgTreeNode [] currentLeafs = new ArgTreeNode[currentLeafsLength];
        for(int i=0; i<newLeafs.length; i++)
            currentLeafs[i] = newLeafs[i];
        for(int i=0; i<oldLeafsLength; i++)        
            currentLeafs[newLeafs.length+i] = oldLeafs[i];
        this.leafs = currentLeafs;
    }
    private void currentRoundLeafs(ArgTreeNode newLeaf){
        boolean[] attacked = new boolean[this.leafs.length];
        for(int i=0; i<this.leafs.length; i++){
            if(this.leafs[i].getPlayer()==newLeaf.getParent().getPlayer())
                attacked[i]=true;
            else
                attacked[i]=false;
        }
        ArgTreeNode[] oldLeafs = this.oldLeafs(attacked);
        int currentLeafsLength;
        if(oldLeafs ==null) currentLeafsLength=1;
        else currentLeafsLength =oldLeafs.length + 1;
        ArgTreeNode [] currentLeafs = new ArgTreeNode[currentLeafsLength];
        currentLeafs[0] = newLeaf;        
        for(int i=1; i<currentLeafsLength; i++)                    
            currentLeafs[i] = oldLeafs[i-1];
        this.leafs = currentLeafs;
    }
    /**return the leaf node matching the parameters*/
    private ArgTreeNode getLeafNode(int source, int target, int level){
        ArgTreeNode leafNode = null;
        boolean found=false;
        int index=0;
        while((!found)&&(index<leafs.length)){
            if((leafs[index].getPlayer()== source)&&(leafs[index].getParent().getPlayer()==target)
            &&(leafs[index].getLevel()==level)){
                found = true;
                leafNode = leafs[index];
            }
            else
                index ++;
        }
        return leafNode;
    }
    /**delete the given leafs*/
    private void deleteLeaf(ArgTreeNode leaf){
        ArgTreeNode[] newLeafs = new ArgTreeNode [leafs.length-1];
        int j=0;
        for(int i=0; i< leafs.length; i++){
            if (!leafs[i].equals(leaf)){
                newLeafs[j]=leafs[i];
                j++;
            }
        }
        leafs = newLeafs;
    }
    /**Update leafs links*/
    private void updateLeafs(ArgTreeNode leafNode, ArgTreeNode newLeafNode){
        boolean found  = false;
        int i=0;
        while((!found)&&(i<leafs.length)){
            if(leafs[i].equals(leafNode)){
                leafs[i] = newLeafNode;
                leafs[i].setAttacks(null);
                found=true;
            }
            else
                i++;
        }
    }  
    private void updateLeafs(ArgTreeNode parent){
        if(parent.getAttacks()!=null){
            for(int j=0; j<parent.getAttacks().length; j++){
                if (parent.getAttacks()[j]!=null){
                    for(int i=0; i<leafs.length; i++){
                        if(leafs[i].equals(parent.getAttacks()[j]))
                            leafs[i].setParent(parent);
                    }
                }
            }
        }
    } 
    private void updateLeafsLinks(ArgTreeNode leafNode, ArgTreeNode grandParent){
        boolean found = false;
        int i=0;
        while((!found)&&(i<leafs.length)){
            if(leafs[i].equals(leafNode)){
                leafs[i].setParent(grandParent);
                found = true;
            }
            else
                i++;
        }
    }    
    /**check if node belongs to leaf or not*/
    private boolean belongsToLeafs(ArgTreeNode node){
        boolean belongs = false;
        int i=0;
        while((!belongs)&&(i<leafs.length)){
            if(leafs[i].equals(node))
                belongs = true;
            else
                i++;
        }
        return belongs;
    }
    /**returns the leafs of the tree rooted with subroot*/
    protected ArgTreeNode[] leafs(ArgTreeNode subRoot){
        ArgTreeNode[] subTreeLeafs = null;
        if(subRoot.isleaf()){
                ArgTreeNode[] nodeList = {subRoot};
                subTreeLeafs =  BasicOperations.append(subTreeLeafs, nodeList);
            }
            else{
                if(subRoot.getAttacks()!=null){
                    for(int i=0; i<subRoot.getAttacks().length; i++){
                        if(subRoot.getAttacks()[i]!=null)                        
                            subTreeLeafs =  BasicOperations.append(subTreeLeafs, leafs(subRoot.getAttacks()[i]));
                    }
                }
            } 
             return subTreeLeafs;
        //}       
    }    
    /**returns the strongest Leaf (not equal node)*/ 
    private ArgTreeNode getStrongestGreenLeaf(ArgTreeNode node){        
        if(!this.hasPossibleGreenLeafs()) return null;
        if(this.leafs.length==1){
            if(!this.leafs[0].equals(node)) return this.leafs[0];
            else return null;
        }
        if(!leafs[0].equals(node)){
            ArgTreeNode strongestLeaf = leafs[0];
            for(int i=1; i<leafs.length; i++){
                if((leafs[i].getConfidence()> strongestLeaf.getConfidence())&&(!leafs[i].equals(node)))
                    strongestLeaf = leafs[i];
            }            
            return strongestLeaf;
        }
        else{
            ArgTreeNode strongestLeaf = leafs[1];
            for(int i=2; i<leafs.length; i++){
                if(leafs[i].getConfidence()> strongestLeaf.getConfidence())
                    strongestLeaf = leafs[i];
            }
            return strongestLeaf;
        }        
    }
    /**returns the strongest Leaf*/ 
    private ArgTreeNode getStrongestGreenLeaf(){        
        if(!this.hasGreenLeafs())   return null;
        if(this.leafs.length==1)    return this.leafs[0];
        else{
           ArgTreeNode strongestLeaf = leafs[0]; 
           for(int i=1; i<leafs.length; i++){
               if(leafs[i].getConfidence()> strongestLeaf.getConfidence())
                   strongestLeaf = leafs[i];
           }
           return strongestLeaf;
        }
    }    
    /***************************************************************************************************************/
    //ADD NEW NODE METHODS    
    /*for the time being target in moves refere to the leaf node index+1*/
    protected boolean addRoundMoves(Move[] moves, int round){        
       boolean[] leafsAttacked = new boolean[leafs.length];
       for(int i=0; i<leafsAttacked.length; i++){leafsAttacked[i]=false;}        
        int movesIndx=0;                
        boolean done = true;
        ArgTreeNode[] newLeafs = new ArgTreeNode[moves.length];
        while((done)&&(movesIndx<moves.length)){ 
            if(this.treeSize + 1 > this.maxTreeSize){done = false;}
            else{
                int colour = MoveType.originalColour(moves[movesIndx].getType()); 
                int parentLeafIndx = moves[movesIndx].getLeafIndx();
                ArgTreeNode parent;
                if(parentLeafIndx!=-1)
                    parent = leafs[parentLeafIndx];
                else
                    parent = leafs[0];
                leafsAttacked[parentLeafIndx]=true;
                int playerIndx = moves[movesIndx].getSource();
                double conf = moves[movesIndx].getConfidence();            
                int type = moves[movesIndx].getType();
                ArgTreeNode treeNode = new ArgTreeNode(playerIndx, conf, type, colour, round, parent);
                //treeNode.setColoured(true);
                newLeafs[movesIndx] = treeNode;
                
                if (parent.getAttacks()==null)
                    parent.initializeAttacks(this.currentNoPlayers);
                parent.addAttack(treeNode);
                parent.setColoured(false);
                this.treeSize++;
               
               if(treeNode.getConfidence()>this.getGreenConf())                    
                    this.greenConfNode = treeNode;                
                movesIndx++;
           }                
        }
        this.currentRoundLeafs(newLeafs,leafsAttacked);
        this.colour();
        return done;
    }
    protected boolean addRoundBestMove(Move move, int round){
        if(this.treeSize+1>this.maxTreeSize) return false;
        else{            
            int colour = MoveType.originalColour(move.getType()); 
            int parentLeafIndx = move.getLeafIndx();
            ArgTreeNode parent;
            if(parentLeafIndx!=-1)
                parent = this.leafs[parentLeafIndx];
            else
                parent = this.leafs[0];
            int playerIndx = move.getSource();
            double conf = move.getConfidence();            
            int type = move.getType();
            ArgTreeNode treeNode = new ArgTreeNode(playerIndx, conf, type, colour, round, parent);
            if (parent.getAttacks()==null)
                parent.initializeAttacks(this.currentNoPlayers);
            parent.addAttack(treeNode);
            parent.setColoured(false);
            this.treeSize++;
            if(treeNode.getConfidence()>this.getGreenConf())
                this.greenConfNode = treeNode;                
            this.currentRoundLeafs(treeNode);
            this.colour();
            return true;
        }
    }        
    protected void addRootMove(Move move){
       ArgTreeNode rootNode = new ArgTreeNode(1, move.getConfidence(),  move.getType(), Colours.GREEN,  1,  null);
       if(this.leafs==null)
           this.leafs= new ArgTreeNode[1];
       this.leafs[0]= rootNode;
       this.leafs[0].setColoured(true);
       this.root = rootNode;
       rootNode.setColoured(true);
       this.treeSize=1;       
       this.greenConfNode = this.root;
       this.activeGreenConfNode = this.root;
    } 
    /***************************************************************************************************************/
     /**coloring methods**/
    /**return the strongest Node which original colour is green, that is not equal to node, and which 
     confidence is larger than startConfidence*/
    private ArgTreeNode getStrongestGreenNode(ArgTreeNode rootNode, double startConfidence){
        ArgTreeNode strongestNode = null;
        if(rootNode!=null){
            if((rootNode.getColour()==Colours.GREEN)&&(rootNode.getConfidence()>startConfidence)){
                strongestNode = rootNode;
                if(rootNode.getAttacks()!=null){
                    for(int i=0; i<rootNode.getAttacks().length; i++){
                        if(rootNode.getAttacks()[i]!=null){
                          ArgTreeNode childStrongestNode =  this.getStrongestGreenNode(rootNode.getAttacks()[i], strongestNode.getConfidence());
                          if(childStrongestNode!=null){
                              if(childStrongestNode.getConfidence()>strongestNode.getConfidence())
                                  strongestNode = childStrongestNode;
                          }
                        }
                    }
                }
            }
            else{
                if(rootNode.getAttacks()!=null){
                   for(int i=0; i<rootNode.getAttacks().length; i++){
                       if(rootNode.getAttacks()[i]!=null){
                           ArgTreeNode childStrongestNode = this.getStrongestGreenNode(rootNode.getAttacks()[i], startConfidence);
                           if (childStrongestNode!=null){
                               if(strongestNode!=null){
                                   if(strongestNode.getConfidence()<childStrongestNode.getConfidence())
                                       strongestNode = childStrongestNode;
                               }
                               else
                                   strongestNode = childStrongestNode;                                                                  
                           }
                       }
                   }
                }
            }
        }
        return strongestNode;
    }
    /**updates the active green confidence*/
    private void activeGreenConfidence(){
        if(this.greenConfNode.getColour()==Colours.GREEN)
            this.activeGreenConfNode = this.greenConfNode;
        else{
            ArgTreeNode strongesGreenLeaf = this.getStrongestGreenLeaf();
            if(strongesGreenLeaf==null)
                this.activeGreenConfNode = this.getStrongestGreenNode(this.root,0);
            else{
                ArgTreeNode strongestGreenNode = this.getStrongestGreenNode(this.root,strongesGreenLeaf.getConfidence());
                if(strongestGreenNode==null) this.activeGreenConfNode = strongesGreenLeaf;
                else  this.activeGreenConfNode=strongestGreenNode;
            }                        
        }        
   } 
    /**colours the leafs*/
    private void colourLeafs(){
       for (int i=0; i<leafs.length; i++){
           if(MoveType.originalColour(leafs[i].getType()) == Colours.GREEN){
               if(this.activeGreenConfNode==null){
                   leafs[i].setColour(Colours.PURPLE);
                   leafs[i].setColoured(true);
                   leafs[i].getParent().setColoured(false);
                   if(leafs[i].getParent().getParent()!=null)
                       leafs[i].getParent().getParent().setColoured(false);
               }
               else{
                   if(leafs[i].getConfidence()<this.activeGreenConfNode.getConfidence()){
                       if (leafs[i].getPlayer()!=this.activeGreenConfNode.getPlayer()){
                           leafs[i].setColour(Colours.PURPLE);
                           leafs[i].setColoured(true);
                           leafs[i].getParent().setColoured(false);
                           if(leafs[i].getParent().getParent()!=null)
                               leafs[i].getParent().getParent().setColoured(false);
                       }
                       else{
                           leafs[i].setColour(Colours.GREEN);
                           leafs[i].setColoured(true);
                           leafs[i].getParent().setColoured(false);
                           if(leafs[i].getParent().getParent()!=null)
                               leafs[i].getParent().getParent().setColoured(false);
                       }
                   }
               }
           }
       }
   }
    /**main colouring algorithm*/
    protected void colour(){
        this.activeGreenConfidence();
        this.colourLeafs();
       for(int i=0; i<leafs.length; i++){
            if(leafs[i]!= null){
                if (leafs[i].getColour()!=Colours.PURPLE)
                    colourToRed(leafs[i].getParent());
                else
                    if(leafs[i].getParent().getActiveAttacks()==1)
                        colourToOriginalColour(leafs[i].getParent());                
            }
       }        
        for(int j=0; j<leafs.length; j++){
            if ((leafs[j]!= null)&&(leafs[j].getColour()!=Colours.PURPLE)){   
                ArgTreeNode grandParent = leafs[j].getParent().getParent();
                if(grandParent!=null)
                    colourToOriginalColour(grandParent);                
            }
        }
    } 
    /**colours attacked nodes --RED*/
    private void colourToRed(ArgTreeNode node){
        if(!node.getColoured()){            
            if(node.getAttacks()!=null){
                boolean toRed = false;
                int i=0;
                while((!toRed)&&(i<node.getAttacks().length)){
                    if(node.getAttacks()[i] != null)  {
                        if ((MoveType.originalColour(node.getAttacks()[i].getType()) == Colours.GREEN)||
                            (MoveType.originalColour(node.getAttacks()[i].getType()) == Colours.BLUE))
                            toRed = true;
                        else i++;
                    }
                    else i++;
                }
                if(toRed){
                    node.setColour(Colours.RED);
                    node.setColoured(true);
                    if(node.getParent()!=null)
                        node.getParent().setColoured(false); 
                }
                else{
                    if((MoveType.originalColour(node.getType())==Colours.GREEN)&&
                            (node.getConfidence()<this.activeGreenConfNode.getConfidence())){
                        node.setColour(Colours.PURPLE);
                        node.setColoured(true);
                        if(node.getParent()!=null)
                            node.getParent().setColoured(false); 
                    }
                }
            }            
        }
    }
    /**recolours defended nodes in their ORIGINAL COLOURS*/
    private void colourToOriginalColour(ArgTreeNode node){
         if(!node.getColoured()){
           boolean recolor = true;
           int i=0;
            while((recolor)&&(i<node.getAttacks().length)){
                if(node.getAttacks()[i]!=null){
                    if((node.getAttacks()[i].getColour()!=Colours.RED)&&
                            (node.getAttacks()[i].getColour()!=Colours.PURPLE))
                        recolor=false;   
                }
                i++;
            }
        if(recolor){
           if((MoveType.originalColour(node.getType())==Colours.GREEN)&&((this.activeGreenConfNode!=null)&&
                   (node.getPlayer()!=this.activeGreenConfNode.getPlayer())&&
                   (node.getConfidence()<this.activeGreenConfNode.getConfidence()))){
               node.setColour(Colours.PURPLE);
               node.setColoured(true);
               if(node.getParent()!=null){
                    node.getParent().setColoured(false);
                    this.colourToOriginalColour(node.getParent());               
                    ArgTreeNode grandParent = node.getParent().getParent();
                    if(grandParent!=null)
                        this.colourToRed(grandParent);
               }
           }
           else{
               node.setColour(MoveType.originalColour(node.getType()));
               node.setColoured(true);
               if(node.getParent()!=null){
                    node.getParent().setColoured(false);
                    colourToRed(node.getParent());               
                    ArgTreeNode grandParent = node.getParent().getParent();
                    if(grandParent!=null)
                        colourToOriginalColour(grandParent);
               }
           }
        }             
      }               
    }
   /***************************************************************************************************************/
   /**pruning methods**/
    /**prune the leafs that belongs to the input player*/
    private void pruneLeafs(int playerIndx){        
        ArgTreeNode [] newLeafs = null;        
        ArgTreeNode[] tempList = new ArgTreeNode[1];
        for(int i=0; i<leafs.length; i++){
            if(leafs[i].getPlayer()==playerIndx){                
                if(this.greenConfNode.equals(leafs[i]))
                    this.updateGreenConfidence();
                if (leafs[i].getParent().getActiveAttacks()==1){
                    ArgTreeNode temp = leafs[i].getParent();
                    temp.setAttacks(null);
                    tempList[0] = temp;
                    newLeafs = BasicOperations.append(newLeafs, tempList);
                }
                else{
                    leafs[i].getParent().deleteAttack(leafs[i].getPlayer()-1);
                    leafs[i] = null;
                }
            }
            else{
                tempList[0] = leafs[i];
                newLeafs = BasicOperations.append(newLeafs, tempList);
            }
            
        }        
        this.leafs = newLeafs;
    }
    /**prune the tree which root (subtreenode) is a blue node*/
    private void pruneBlueSubTree(ArgTreeNode subtreeNode, ArgTreeNode parent){
        //get the leafs of the subtee
        ArgTreeNode [] subTreeLeafs = leafs(subtreeNode);
        for(int i=0; i<subTreeLeafs.length; i++)
        {
             //delete those leafs
            this.deleteLeaf(subTreeLeafs[i]);
        }
        //delete the entry in the parent
        parent.deleteAttack(subtreeNode.getPlayer());
    }            
    private void pruneSubTree(int playerIndx, ArgTreeNode subRoot){
        if(subRoot.getPlayer()==playerIndx){
            //delete node .. and replace it with one of its children
        }
        else{
            if(subRoot.getAttacks()!=null){
                for(int i=0; i<subRoot.getAttacks().length; i++){
                    if(subRoot.getAttacks()[i]!=null){
                        if(subRoot.getAttacks()[i].getPlayer() == playerIndx){
                            System.out.print("Found ");
                            System.out.println("Level" + subRoot.getAttacks()[i].getLevel());
                            pruneNode(subRoot.getAttacks()[i], subRoot);
                        }
                        else{
                            pruneSubTree(playerIndx, subRoot.getAttacks()[i]);
                        }
                    }
                }
            }
        }
    }           
    /**deletes the node, and updates the tree links*/
    private void pruneNode (ArgTreeNode node, ArgTreeNode parent){
        int playerIndx = node.getPlayer();
        if(node.getAttacks()!=null){
            for(int i=0; i<node.getAttacks().length; i++){
                if(node.getAttacks()[i] != null){
                    int color = MoveType.originalColour(node.getAttacks()[i].getType());
                    if(color == Colours.BLUE)
                        pruneBlueSubTree(node.getAttacks()[i], node);
                    else
                        this.pruneGreenChild(node.getAttacks()[i], node, node.getParent());                    
                }
            }
            parent = node.getParent();
            if(node.equals(this.greenConfNode))
                this.updateGreenConfidence();            
            parent.deleteAttack(node.getPlayer()-1);
        }        
    }
    /**return the strongest Node which original colour is green, that is not equal to node, and which 
     confidence is larger than startConfidence*/
    private ArgTreeNode getStrongestNode(ArgTreeNode rootNode, ArgTreeNode node, double startConfidence){
        ArgTreeNode strongestNode = null;
        if(rootNode!=null){
            if((!rootNode.equals(node))&&(rootNode.getOriginalColour()==Colours.GREEN)
                &&(rootNode.getConfidence()>startConfidence)){
                strongestNode = rootNode;
                if(rootNode.getAttacks()!=null){
                    for(int i=0; i<rootNode.getAttacks().length; i++){
                        if(rootNode.getAttacks()[i]!=null){
                          ArgTreeNode childStrongestNode =  this.getStrongestNode(rootNode.getAttacks()[i], node, strongestNode.getConfidence());
                          if(childStrongestNode!=null){
                              if(childStrongestNode.getConfidence()>strongestNode.getConfidence())
                                  strongestNode = childStrongestNode;
                          }
                        }
                    }
                }
            }
            else{
                if(rootNode.getAttacks()!=null){
                   for(int i=0; i<rootNode.getAttacks().length; i++){
                       if(rootNode.getAttacks()[i]!=null){
                           ArgTreeNode childStrongestNode = this.getStrongestNode(rootNode.getAttacks()[i], node, startConfidence);
                           if (childStrongestNode!=null){
                               if(strongestNode!=null){
                                   if(strongestNode.getConfidence()<childStrongestNode.getConfidence())
                                       strongestNode = childStrongestNode;
                               }
                               else
                                   strongestNode = childStrongestNode;                                                                  
                           }
                       }
                   }
                }
            }
        }
        return strongestNode;
    }
    /**updates the green confidence*/
    private void updateGreenConfidence(){
        ArgTreeNode strongestLeaf = this.getStrongestGreenLeaf(this.greenConfNode);
        //strongest leaf is the best candidate - now we search the tree in comparing with this candidate
        if(strongestLeaf==null) this.activeGreenConfNode = this.getStrongestNode(this.root, this.greenConfNode,0);
        else{
            ArgTreeNode newGreenConfidence = this.getStrongestNode(this.root, this.greenConfNode,strongestLeaf.getConfidence());
            if(newGreenConfidence==null) this.greenConfNode = strongestLeaf;
            else this.greenConfNode =  newGreenConfidence;
        }
    }
    /**prune the tree which root (child) is a green node*/
    private void pruneGreenChild(ArgTreeNode child, ArgTreeNode node, ArgTreeNode parent){
        ArgTreeNode newNode;
        pruneRoot(node.getPlayer(), child);
        /*if (temp!=null)
            parent = temp;*/
        child = node.getAttacks()[child.getPlayer()-1];
        if (this.greenConfNode.equals(child))
        {
            updateGreenConfidence();
        }
        node.deleteAttack(child.getPlayer()-1);
        if(parent.getPlayer() == child.getPlayer())
        {
             if(parent.getConfidence()<child.getConfidence())
             {
                 newNode = parent.replaceBy(child);
                 newNode.updateAttacks(child.getAttacks());
                 if(child.isleaf())
                 {
                     if(parent.getActiveAttacks()==1)
                     {
                        if(this.greenConfNode.equals(parent))
                         this.greenConfNode = newNode;
                        updateLeafs(child, newNode);
                        parent = newNode;
                     }
                     else
                     {
                         if(this.greenConfNode.equals(parent))
                            this.greenConfNode = newNode;
                         node.setParent(newNode);
                         parent=newNode;
                         this.deleteLeaf(child);                         
                     }
                 }
                 else
                 {
                     if(this.greenConfNode.equals(parent))
                         this.greenConfNode = newNode;
                     node.setParent(newNode);
                     parent=newNode;
                 }
                 
                 if (parent.getParent()!=null)
                 {
                     parent.getParent().updateAttacks(newNode);
                     updateLeafs(parent);
                 }
                 else//root
                 {                     
                     root = parent;
                     updateLeafs(root);
                 }                                  
             }             
        }
        else
        {
          if(child.isleaf())
          {
             updateLeafsLinks(child, parent);             
          }                 
          //parent.updateAttacks(child.getAttacks());        
          node.deleteAttack(child.getPlayer()-1);
          if(parent.equals(this.greenConfNode))
          {
              this.greenConfNode = parent;
          }
        }
       // parent.deleteAttack(node.getPlayer()-1);      
    }
    /**update tree links after pruning*/
    private void updateLinks (ArgTreeNode subRoot){
        if(subRoot!=null){
            if(subRoot.getAttacks()==null)
                updateLeafsLinks(subRoot, subRoot.getParent());
            else{
                for(int i=0; i<subRoot.getAttacks().length; i++){
                    if(subRoot.getAttacks()[i]!=null)
                       subRoot.getAttacks()[i].setParent(subRoot) ;
                }
            }
        }
    }   
    /**apply pruning algorithms to tree rooted at rootNode*/
    private void pruneRoot(int playerIndx, ArgTreeNode rootNode){
        if(playerIndx == rootNode.getPlayer()){
            if(rootNode.getAttacks()!=null){
                for(int i=0; i<rootNode.getAttacks().length; i++){
                    if(rootNode.getAttacks()[i]!=null)
                        pruneRoot(playerIndx, rootNode.getAttacks()[i]);
                }
            }
            ArgTreeNode newRootNode = rootNode.getBestGreenChild();
            if (newRootNode != null){
                if(newRootNode.getAttacks()!=null){
                    for(int i=0; i<newRootNode.getAttacks().length; i++){
                        if((i!=newRootNode.getPlayer()-1)&&(newRootNode.getAttacks()[i]==null)){
                            if(rootNode.getAttacks()[i]!=null)
                                newRootNode.getAttacks()[i] = rootNode.getAttacks()[i];
                        }
                    }
                }
            }
            else
                newRootNode = this.greenConfNode;            
            
            if(this.root.equals(rootNode))
                this.root = newRootNode;
        }
        else{
            if(rootNode.getAttacks()!=null){
                for(int i=0; i< rootNode.getAttacks().length; i++){
                    if(rootNode.getAttacks()[i]!=null){
                        if(rootNode.getAttacks()[i].getPlayer() == playerIndx)
                            this.pruneNode(rootNode.getAttacks()[i], rootNode);                        
                        else
                            this.pruneRoot(playerIndx, rootNode.getAttacks()[i]);
                    }//else leaf
                }
            }
        }        
    }
    /**main pruning algorithm, prune all nodes beloning to the input player*/
    protected void prune(int playerIndx){
       this.pruneLeafs(playerIndx);
       this.pruneRoot(playerIndx,root);
       //this.resetLeafs();
       this.leafs = this.leafs(this.root);       
       this.colour();
       this.currentNoPlayers--;       
    }
    
    /***************************************************************************************************************/
     /**output methods**/  
    private void outPutRoot(ArgTreeNode rootNode, String tab){       
       System.out.println(tab + "P"+rootNode.getPlayer()+" - (" + Colours.translate(rootNode.getColour())+") - ("+rootNode.getConfidence()+")");
       if(rootNode.getAttacks()!=null){
           System.out.println(tab + "attacks");
           for(int i=0; i<rootNode.getAttacks().length; i++){
               if(rootNode.getAttacks()[i]!=null){
                  outPutRoot(rootNode.getAttacks()[i], tab+"\t");
               }               
           }           
       }
    }      
    protected void outPut(){this.outPutRoot(this.root, "");}
    /***************************************************************************************************************/
     /*Managing Requests from Tree Oriented Players*/
    /**1 return the status of the Green Confidence i.e. Green or Attacked*/
    protected boolean greenConfStatus(){
        if(this.greenConfNode.getColour()==Colours.GREEN)   return true;
        else    return false;
    }
    protected int greenConfplayer(){return this.greenConfNode.getPlayer();}
    
    /**2 return the Green Confidence SubTree Leafs*/
    private int getLeafIndex(ArgTreeNode leafNode){
        if(leafs==null)
            return -2;
        else{
            for(int i=0; i<leafs.length; i++){
                if(leafs[i].equals(leafNode))
                    return i;
            }
            return -1;//leaf not found
        }
    }
    private ArgTreeNode[] greenConfLeafs(){
        ArgTreeNode greenConfRoot = this.greenConfNode;
        if(greenConfRoot.isleaf()){
            ArgTreeNode[] greenConfLeafs = new ArgTreeNode[1];
            greenConfLeafs[0] = greenConfRoot;
            return greenConfLeafs;
        }
        else//Green Confidence Node is not 
        {
            return this.leafs(greenConfRoot);             
        }
    }
    protected CompressedArgTreeNode[] greenConfCompressedLeafs(int focusingLevel){        
        return this.getCompressedLeafs(focusingLevel, this.greenConfLeafs());    
    }
    
    /**3 return the status of the Active Green Confidence i.e. Green or Attacked*/
    protected boolean activeGreenConfStatus(){
        if(this.activeGreenConfNode!=null){
            if(this.activeGreenConfNode.getColour()==Colours.GREEN)
                return true;
            else
                return false;
        }
        else
            return false;
    }
    protected int activeGreenNodePlayer(){
        if(this.activeGreenConfNode!=null) return this.activeGreenConfNode.getPlayer();
        else    return -1;
    }
    protected boolean activeGreenEqualsGreenConf(){
        if(this.activeGreenConfNode==null) return false;
        if(this.activeGreenConfNode.equals(this.greenConfNode))
            return true;
        else return false;
    }
    
    /**4 return the Active Green Confidence SubTree Leafs*/
    private ArgTreeNode[] activeGreenConfLeafs(){
        ArgTreeNode activeGreenConfRoot = this.activeGreenConfNode;
        if(activeGreenConfRoot.isleaf()){
            ArgTreeNode[] activeGreenConfLeafs = new ArgTreeNode[1];
            activeGreenConfLeafs[0] = activeGreenConfRoot;
            return activeGreenConfLeafs;
        }
        else//Green Confidence Node is not 
        {
            return this.leafs(activeGreenConfRoot);             
        }
    }
    protected CompressedArgTreeNode[] activeGreenConfCompressedLeafs(int focusingLevel){        
        return this.getCompressedLeafs(focusingLevel, this.activeGreenConfLeafs());
    }
    
    /**5 find the Blue Dominant Player*/
    protected int[] countBlueNodes(){//return the index number of the dominant blue player
        int[] blueNodesCounter = new int[this.currentNoPlayers];
        for (int i=0;i<blueNodesCounter.length;i++)
            blueNodesCounter[i]=0;
        blueNodesCounter = this.countBlueNodes(this.root, blueNodesCounter);
        return blueNodesCounter;
    }
    private int[] countBlueNodes(ArgTreeNode rootNode, int[]blueCounter){
        int[] currentCounter = new int[blueCounter.length];
        for(int i=0; i<currentCounter.length; i++)
            currentCounter[i]=blueCounter[i];
        if(rootNode.getColour()==Colours.BLUE)
            currentCounter[rootNode.getPlayer()-1]++;
        if(rootNode.getAttacks()!=null){
            for(int i=0; i<rootNode.getAttacks().length; i++){
                if(rootNode.getAttacks()[i]!=null){
                    ArgTreeNode child = rootNode.getAttacks()[i];
                    int[] childCounter= this.countBlueNodes(child, currentCounter);
                    for(int j=0; j<childCounter.length; j++)
                        currentCounter[j] = currentCounter[j] + childCounter[j];
                }
            }
        }
        return currentCounter;
    }
    
    /**6 return the leafs indirectly attacking player P*/
    /*for each leaf traverse back to the root, 
     if a (red) P-Node was found then the leaf is an indirect threat*/
    private ArgTreeNode[] indirectThreats(int playerIndx){        
        if(this.leafs==null)
            return null;
        else{
            ArgTreeNode[] indirectThreat = null;
            for(int i =0; i<leafs.length; i++){
                if(leafs[i]!=null){
                    if ((leafs[i].getParent()!=null)&&(leafs[i].getPlayer()!=playerIndx)
                    &&(leafs[i].getParent().getPlayer()!=playerIndx)){                       
                        if(this.indirectThreat(playerIndx, leafs[i].getParent())){
                            ArgTreeNode[] newThreat = new ArgTreeNode[1];
                            newThreat[0] = leafs[i];
                            BasicOperations.append(indirectThreat, newThreat);
                        }                       
                    }
                }
            }
            return indirectThreat;   
        }        
    }
    private boolean indirectThreat(int playerIndx, ArgTreeNode node){
        if(node.getParent()!=null){
            if(node.getParent().getPlayer()==playerIndx)
                return true;
            else
                return indirectThreat(playerIndx,node.getParent());
        }
        else
            return false;
    }
    protected CompressedArgTreeNode[] compressedIndirectThreats(int playerIndx, int focusingLevel){
        ArgTreeNode[] indirectThreat = this.indirectThreats(playerIndx);
        return this.getCompressedLeafs(focusingLevel, indirectThreat);
    }
    
    /*Attack When Needed Methods*/    
    protected boolean hasGreenNodes(int playerIndx){
        boolean hasTreeNodes = this.hasGreenLeafs(playerIndx);
        if(hasTreeNodes)
            return true;
        else
        return this.hasGreenNodes(playerIndx, this.root);
    }
    private boolean hasGreenLeafs(int playerIndx){        
        if(this.leafs==null)
            return false;
        else{
            for(int i=0; i<this.leafs.length; i++){
                if((this.leafs[i].getPlayer()==playerIndx)&&(this.leafs[i].getColour()==Colours.GREEN))
                    return true;
            }
            return false;
        }        
    }
    private boolean hasGreenNodes(int playerIndx, ArgTreeNode rootNode){
        if(rootNode == null){
            return false;
        }
        else{
            if((rootNode.getPlayer()==playerIndx)&&(rootNode.getColour()==Colours.GREEN))
                return true;
            else{
                if(rootNode.getAttacks()==null)
                    return false;
                else{
                    for(int i=0; i<rootNode.getAttacks().length; i++){
                        if(rootNode.getAttacks()[i]!=null){
                            boolean done = this.hasGreenNodes(playerIndx, rootNode.getAttacks()[i]);
                            if(done) return true;
                        }
                    }
                    return false;
                }
            }                
        }            
    }
    protected int dominantBlue(){
        int[] blueCounter = this.countBlueNodes();
        int dominantBlue = 0;
        for(int i=1; i<blueCounter.length; i++){
            if(blueCounter[i]>blueCounter[dominantBlue])
                dominantBlue = i;
        }
        if(blueCounter[dominantBlue]>0) return dominantBlue+1;
        else return -1;
    }
    protected int[] dominantBlues(){
        int[] blueCounter = this.countBlueNodes();
        int[] dominantBlues = null;
        int domIndx=0;
        while((domIndx<blueCounter.length)&&(blueCounter[domIndx]==0))
            domIndx++;
        if(domIndx<blueCounter.length){
            int maxBlue = blueCounter[domIndx];
            for(int i=domIndx+1; i<blueCounter.length; i++){
                if(blueCounter[i]>maxBlue)
                    maxBlue = blueCounter[i];
            }
            for(int i=0; i<blueCounter.length; i++){
                if(blueCounter[i]==maxBlue)
                    dominantBlues=BasicOperations.append(dominantBlues, i+1);
            }
            return dominantBlues;
        }
        else return null;
    }   
    protected int dominantExclusiveBlue(){
        int[] blueCounter = this.countBlueNodes();
        if(blueCounter==null) return -2; //no blue nodes
        else{
            int i=0;
            while((blueCounter[i]==0)&&(i<blueCounter.length)) i++;
            if(i<blueCounter.length){
                int dominantBLue = i;
                i=i+1;
                for(;i<blueCounter.length; i++){
                    if(blueCounter[i]>0)
                        return -1;
                }
                return dominantBLue+1;
            }
            else return -2;//no blue nodes
        }
    }
    private ArgTreeNode[] playerLeafs(int playerIndx){
        ArgTreeNode[] playerLeafs = null;
        for(int i=0; i<leafs.length; i++){
            if(leafs[i].getPlayer()==playerIndx){
                ArgTreeNode[] playerLeaf = new ArgTreeNode[1];
                playerLeaf[0] = leafs[i];
                playerLeafs = BasicOperations.append(playerLeafs,playerLeaf);
            }
        }
        return playerLeafs; 
    }
    protected CompressedArgTreeNode[] compressedPlayerLeafs(int playerIndx, int focusingLevel){
        ArgTreeNode[] playerLeafs = this.playerLeafs(playerIndx);
        return this.getCompressedLeafs(focusingLevel, playerLeafs);
    }    
    /*-------------------------------------------------------------------------------------------*/
}//end of the class