/*
 * ArgTreeNode.java
 *
 * Created on 02 September 2008, 17:19
 *
 * represents the nodes of the argumentation tree, and their basic operations
 * @author maya
 */

package GUIComponents;

public class ArgTreeNode {
    private int player; //player index: represents the index of the player who played the move represented by the node
    private double confidence;//the confidence of the move represented by the node
    private int colour;//the node colour
    private int type; //the type of the moverepresented by the node
    private int level;//i.e. the round in which the game was played
    private ArgTreeNode[] attacks;//links to the nodes attacking this node
    private ArgTreeNode parent;//link to the node parent
    private boolean coloured;//set to true when the node is (re)coloured after each round

    //constructors
    public ArgTreeNode(int player, double conf, int type, int colour, int level, ArgTreeNode parent) {
        this.player = player;
        this.confidence = conf;
        this.type=type;
        this.colour = colour;
        this.coloured=false;
        this.level = level;
        this.parent = parent;
        this.attacks=null;
    }
    public ArgTreeNode(ArgTreeNode node){
        this.player = node.getPlayer();
        this.confidence = node.getConfidence();
        this.type=node.getType();
        this.colour = node.getColour();
        this.coloured=node.getColoured();
        this.level = node.getLevel();
        this.parent = node.getParent();
        this.attacks=node.getAttacks();
    }
            
    /**Setters & Getters*/
    protected int getPlayer() {return this.player;}
    protected double getConfidence() {return this.confidence;}
    protected int getColour() {return this.colour;}
    protected int getOriginalColour(){
        switch (this.type){
            case MoveType.DISTINGUISH: case MoveType.UNWANTED_CONSEQUENCES:
                return Colours.BLUE;
            default:
                return Colours.GREEN;
        }
    }
    protected int getType() {return this.type;}
    protected int getLevel() {return this.level;}
    protected ArgTreeNode getParent(){return this.parent;}    
    protected ArgTreeNode[] getAttacks() {return this.attacks;}
    protected boolean getColoured(){return this.coloured;}
    
    protected void setColour(int newColour){this.colour = newColour;}
    protected void setColoured(boolean coloured){this.coloured = coloured;}
    protected void setAttacks(ArgTreeNode[] newAttacks) {this.attacks = newAttacks;}
    protected void setParent(ArgTreeNode newParent){this.parent = newParent;}
    
    //METHODS
    /**add a new attack against this node*/
    protected void addAttack(ArgTreeNode attack){attacks[attack.getPlayer()-1] = attack;}
    /**initializes the attacks array*/
    protected void initializeAttacks(int maxPlayer){this.attacks = new ArgTreeNode[maxPlayer];}
    /**return true if this node is the root (i.e. has no parent)*/
    protected boolean isRoot(){
        if(this.parent == null) return true;
        else    return false;
    }
    /**return true if this node is a leaf node (i.e. has no attacks/children)*/
    protected boolean isleaf(){
        if(this.attacks == null) return true;
        else    return false;
    }
    /**replaces a given node with a new copy*/
    protected ArgTreeNode replaceBy(ArgTreeNode node){
        ArgTreeNode newNode = new ArgTreeNode(this.player, node.getConfidence(), node.getType(), node.getColour(), node.getLevel(), this.parent);
        newNode.setColoured(false);
        newNode.setAttacks(this.attacks);
        newNode.updateAttacks(node.getAttacks());   
        return newNode;
    }
    /**replaces the attacks against the node with a new set of attacks*/
    protected void updateAttacks(ArgTreeNode[] newAttacks){
        if(newAttacks != null){
            for(int i=0; i<newAttacks.length; i++){
                if(newAttacks[i]!=null){
                    attacks[i]= newAttacks[i];
                    newAttacks[i].setParent(this);
                }
            }
        }
    }    
    protected void updateAttacks(ArgTreeNode newAttack){
        attacks[newAttack.getPlayer()-1]= newAttack;
        newAttack.setParent(this);
    }
    /**delete a given attack against this node*/
    protected void deleteAttack(int attackIndex){
        attacks[attackIndex]= null;
        /*if(this.getActiveAttacks()==0)
            this.attacks =null;*/
    }
    /**returns true if this node is equal to the given node*/
    protected boolean equals(ArgTreeNode node){
        if((this.player == node.getPlayer())&&(this.level == node.getLevel())){
            if(node.getParent()==null){
                if (this.getParent()==null) return true;
                else    return false;           
            }
            else{
                if(this.getParent()==null)  return false;                     
                else{
                    if(this.getParent().getPlayer() == node.getParent().getPlayer())
                        return true;
                    else    return false; 
                }
            }
        }
        else
            return false;
    }
    /**returns true if this node is in the given nodes list*/
    protected boolean in(ArgTreeNode[] nodesList){
        if(nodesList==null)return false;
        for(int i=0; i<nodesList.length; i++){
            if(this.equals(nodesList[i]))
                return true;
        }
        return false;
    }
    /**returns the number of active attacks (non null)*/
    protected int getActiveAttacks(){
        int counter=0;
        if(this.attacks!=null){
            for(int i=0; i<this.attacks.length; i++){
                if(this.attacks[i]!=null)   counter++;
            }
        }
        return counter;
    }

    /**return the highest confidence Green Attack against this node*/
    protected ArgTreeNode getBestGreenChild(){
        ArgTreeNode bestGreenChild = null;
        if(this.attacks!=null){
            for(int i=0; i<attacks.length; i++){
                if((attacks[i]!=null)&&(MoveType.originalColour(attacks[i].getType())==Colours.GREEN)){
                    if(bestGreenChild==null)                    
                        bestGreenChild = attacks[i];                    
                    else{
                        if(bestGreenChild.getConfidence()<attacks[i].getConfidence())
                            bestGreenChild = attacks[i];
                    }
                }
            }
        }
        return bestGreenChild;
    }
   /**return true if the node is attacked by the given attack*/
    protected boolean attackedBy(ArgTreeNode attack){
        if(attack==null) return false;
        if(this.attacks==null) return false;
        for(int i=0; i<this.attacks.length; i++){
            if((this.attacks[i]!=null)&&(this.attacks[i].equals(attack)))
                return true;
        }
        return false;
    }
    /**return true if the node is attacked by the given set of attacks*/
    protected boolean attackedBy(ArgTreeNode[] attacks){
        if(attacks==null) return false;
        else{
            for(int i=0; i<attacks.length; i++){
                if(this.attackedBy(attacks[i])) return true;
            }                
        }
        return false;
    }
    
    /**return the inetersection result of the two input ArgTreeNode lists*/
    protected static ArgTreeNode[] intersects(ArgTreeNode[] nodeList1, ArgTreeNode[] nodeList2){
        if(nodeList1==null) return null;
        if(nodeList2==null) return null;
        ArgTreeNode[] intersect = null;
       for(int i=0; i<nodeList1.length; i++){
            for(int j=0; j<nodeList2.length; j++){
                if(nodeList1[i].equals(nodeList2[j]))
                    intersect = BasicOperations.append(intersect,nodeList1[i]);
            }
        }
        return intersect;
    }
}//End of the Class

