package GUIComponents;
/* Class structure

AssocRuleMining
      |
      +-- RuleList			*/

// Java packages
import java.io.*;
import java.util.*;


/** Set of utilities to support various Association Rule Mining (ARM) 
algorithms included in the LUCS-KDD suite of ARM programs. 
@author Frans Coenen
@version 2 March 2004
@ modified by Maya Wardeh for PADUA system 2007
@ modified by Maya Wardeh for PISA  system 2008*/

public class RuleList extends AssocRuleMining {
    /* ------ FIELDS ------ */    
    protected RuleNode startRulelist = null;//The reference to start of the rule list
    
    /** Default constructor to create an instance of the class RuleList  */    	
    public RuleList(){}

    /* ------ GETTERS ------ */
    /*returns the first rule in the list**/
    protected RuleNode getFirstRule() {return startRulelist;}
    /*cuts and returns the first rule in the list **/
    protected RuleNode cutFirstRule(){
       RuleNode firstRule = startRulelist;
       startRulelist = startRulelist.next;
       return firstRule;
    }
    /*returns the last rule in the list**/
    protected RuleNode getLastRule(){ 
        RuleNode linkRuleNode =  this.startRulelist;
        if(linkRuleNode !=null)
            while (linkRuleNode.next != null)
                linkRuleNode = linkRuleNode.next;
        return linkRuleNode;
    }

    /* ------ METHODS ------ */	
    /* ------ INSERTS ------ */	
    /* Methods for inserting rules into a linked list of rules ordered 
    according to confidence (most confident first). Each rule descibed in 
    terms of 3 fields: 1) Antecedent (an ite set), 2) a consequent (an item 
    set), 3) a confidence value (double). <P> The suppoert field is not used. */      
    /** Inserts an (association/classification) rule into the linkedlist of 
    rules pointed at by <TT>startRulelist</TT>. <P> The list is ordered so that
    rules with highest confidence are listed first. If two rules have the same
    confidence the new rule will be placed after the existing rule. Thus, if
    using an Apriori approach to generating rules, more general rules will
    appear first in the list with more specifuc rules (i.e. rules with a larger
    antecedent) appearing later as the more general rules will be generated
    first.   
    @param antecedent the antecedent (LHS) of the rule.
    @param consequent the consequent (RHS) of the rule.
    @param supportForRule the associated support value.  */    
    protected void insertRuleintoRulelist(short[] antecedent, 
    				short[] consequent, double confidenceForRule) {  

	// Create new node
	RuleNode newNode = new RuleNode(antecedent,consequent,
							confidenceForRule);
	  
	// Empty list situation
	if (startRulelist == null) {
	    startRulelist = newNode;
	    return;
	    }
		
	// Add new node to start	
	if (confidenceForRule > startRulelist.confidenceForRule) {
	    newNode.next = startRulelist;
	    startRulelist  = newNode;
	    return;
	    }
	
	// Add new node to middle
	RuleNode markerNode = startRulelist;
	RuleNode linkRuleNode = startRulelist.next;
	while (linkRuleNode != null) {
	    if (confidenceForRule > linkRuleNode.confidenceForRule) {
	        markerNode.next = newNode;
		newNode.next    = linkRuleNode;
		return;
		}
	    markerNode = linkRuleNode;
	    linkRuleNode = linkRuleNode.next;	    
	    }  
	
	// Add new node to end
	markerNode.next = newNode;
	}
    
    protected void insertRuleintoRulelist(RuleNode newNode) {  	  
	// Empty list situation
	if (startRulelist == null) {
	    startRulelist = newNode;
	    return;
	    }
		
	// Add new node to start	
	if (newNode.getConfidence() > startRulelist.confidenceForRule) {
	    newNode.next = startRulelist;
	    startRulelist  = newNode;
	    return;
	    }
	
	// Add new node to middle
	RuleNode markerNode = startRulelist;
	RuleNode linkRuleNode = startRulelist.next;
	while (linkRuleNode != null) {
	    if (newNode.getConfidence() > linkRuleNode.confidenceForRule) {
	        markerNode.next = newNode;
		newNode.next    = linkRuleNode;
		return;
		}
	    markerNode = linkRuleNode;
	    linkRuleNode = linkRuleNode.next;	    
	    }  
	
	// Add new node to end
	markerNode.next = newNode;
	}
    
    /* ------ GETTERS ------ */		
    /* GET NUMBER OF RULES */
    public int getNumCRs() {
        int number = 0;
        RuleNode linkRuleNode = startRulelist;
	
	// Loop through linked list
	while (linkRuleNode != null) {
	    number++;
	    linkRuleNode = linkRuleNode.next;
	    }
	
	// Return
	return(number);
	}


    /* ------ SETTERS ------ */	
    /* SET RECONVERSION ARRAYS */
    /** Sets the reconversion array reference values.
    @param conversionArrayRef the reference to the 2-D array used to renumber
    coulmns for input data in terms of frequency of single attributes
    (reordering will enhance performance for some ARM and CARM algorithms).
    @param reconversionArrayRef the reference to the 1-D array used to reconvert
    input data column numbers to their original numbering where the input data
    has been ordered to enhance computational efficienvy. */
    protected void setReconversionArrayRefs(int[][] conversionArrayRef,
    					short[] reconversionArrayRef) {
        conversionArray   = conversionArrayRef;
        reconversionArray = reconversionArrayRef;
        }
        	
    /* ------ OUTPUTS ------ */	
    /* OUTPUT RULE LINKED LIST */
    /** Outputs contents of rule linked list (if any) */   
    public void outputRules() {
        outputRules(startRulelist);
	}	  
    private void outputRules(RuleNode ruleList) {
	// Check for empty rule list
	if (ruleList==null) System.out.println("No rules generated!");
	
	// Loop through rule list
	int number = 1;
        RuleNode linkRuleNode = ruleList;
	while (linkRuleNode != null) {
	    System.out.print("(" + number + ") ");
	    outputRule(linkRuleNode);
            System.out.println(" " + linkRuleNode.confidenceForRule + "%");
	    number++;
	    linkRuleNode = linkRuleNode.next;
	    }
	}
    
    /** Outputs a rule.
    @param rule the rule to be output. */
    private void outputRule(RuleNode rule) {
        outputItemSet(rule.antecedent);
	System.out.print(" -> ");
        outputItemSet(rule.consequent);
	}
    
    /* OUTPUT RULE LINKED LIST WITH RECONVERSION */   
    protected void outputRulesWithReconversion() {
	// Check for empty rule list
	if (startRulelist==null) System.out.println("No rules generated!");
	
	// Loop through rule list
        int number = 1;
        RuleNode linkRuleNode = startRulelist;
	while (linkRuleNode != null) {
	    System.out.print("(" + number + ") ");
	    outputItemSetWithReconversion(linkRuleNode.antecedent);
	    System.out.print(" -> ");
            outputItemSetWithReconversion(linkRuleNode.consequent);
            System.out.println(" " + linkRuleNode.confidenceForRule + "%");
	    number++;
	    linkRuleNode = linkRuleNode.next;
	    }
	}

    /* OUTPUT NUMBER OF RULES */    
    protected void outputNumRules() {
        System.out.println("Number of rules         = " + getNumCRs());
	}
   
   
   /******************************************************************************/
   //New Methods
   /**merges the current list with 'newList'*/
   protected void merge(RuleList newList){
      if(newList != null)
      {
        RuleNode linkRuleNode = newList.getFirstRule();
        while(linkRuleNode!=null)
        {
          short [] antecedent = linkRuleNode.getAntecednt();
          short [] consequent = linkRuleNode.getConsequent();
          double   confidence = linkRuleNode.getConfidence();
	  this.insertRuleintoRulelist(antecedent,consequent, confidence);
          linkRuleNode = linkRuleNode.next;
        }
      }
   }
   
   /*returns the best rule (i.e. of highest confidence -- win mode) and higher than confidence    
     threshold*/
   protected RuleNode getBestRuleOfConfidence(double confidenceThreshold){
      RuleNode linkRuleNode = this.startRulelist;
      RuleNode bestRule = null;
      while(linkRuleNode!=null){
        double   confidence = linkRuleNode.getConfidence();
        if(confidence >= confidenceThreshold){          
          if(bestRule == null){
            bestRule = new RuleNode(linkRuleNode.getAntecednt(), 
                                    linkRuleNode.getConsequent(), confidence);
          }
          else
            if(confidence > bestRule.getConfidence()){
               bestRule = new RuleNode(linkRuleNode.getAntecednt(), 
                                    linkRuleNode.getConsequent(), confidence);
            }
        }
        linkRuleNode = linkRuleNode.next;
      }
      return bestRule;
   }   
        
   /**returs true if Rule exists in the current list*/   
   protected boolean existsInThisList(RuleNode rule){
       return this.existsInThisList(rule, this.startRulelist);
   }
   private boolean existsInThisList(RuleNode rule, RuleNode startNode){
       if(startNode==null){return false;}
       else{
           if(startNode.getConfidence()>rule.getConfidence())   return false;
           else{
              while((startNode!=null)&&
                      (startNode.getConfidence()<=rule.getConfidence())){
                  if(RuleNode.equals(rule,startNode))   return true;
                  else      startNode = startNode.next;  
              }
              return false;
           }
       }
   }   
}//end of the class
